/*
 * Copyright (c) 2010, Aragon Gouveia <aragon@phat.za.net>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials provided
 *   with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* FORMAT STRINGS

%bl         Battery level, integer, percent
%bt         Battery time
%bsm        Battery state max
%bsd        Battery state discharging
%bsc        Battery state charging
%bsC        Battery state critical
%bsD        Battery state disconnected
%cl         CPU usage, integer, percent
%cc         CPU clock, integer, hertz

*/

#define STATF_VERSION "0.0.1"

#include <sys/param.h>
#include <sys/errno.h>
#include <sys/resource.h>
#include <sys/sysctl.h>
#include <sys/ioctl.h>
#include <dev/acpica/acpiio.h>

#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sysexits.h>
#include <fcntl.h>
#include <unistd.h>
#include <devstat.h>
#include <math.h>
#ifdef WITH_DWM
#include <X11/Xlib.h>
#endif

/* platform specific defines */
#define ACPIDEV "/dev/acpi"
#define VLIST_PRIO_GEN 1
#define VLIST_PRIO_ACPIIOCTL 2
#define VLIST_PRIO_BIOREAD 3
#define BSTATE_CHARGING 1
#define BSTATE_DISCHARGING 2
#define BSTATE_CRITICAL 3
#define BSTATE_DISCONNECTED 4
#define BSTATE_MAX 5

/* scrotbar defines */
#define FORMAT_MAX 128
#define FSTRING_STATE_ROOT 0
#define FSTRING_STATE_FMT 1
#define INFO_SET_BL 1
#define INFO_SET_BSM 2
#define INFO_SET_BSD 3
#define INFO_SET_BSC 4
#define INFO_SET_BSG 5
#define INFO_SET_BSA 6
#define INFO_SET_BT 7
#define INFO_SET_CL 8
#define INFO_SET_CC 9

typedef struct output_t {
	char *msgpart;
	uint_fast8_t conditional;
	struct output_t *jump;
	struct output_t *next;
} output_t;
typedef struct format_t {
	output_t *output;
	int_fast8_t padding;
} format_t;
typedef struct battall_t { /* global battery data */
	int acpifd;
	union acpi_battery_ioctl_arg battio;
	const char *pwr_units;
} battall_t;
typedef struct batt_t { /* battery data */
	int formatc;
	format_t *format;
	const char *units;
} batt_t;
typedef struct cpu_t { /* CPU data */
	int formatc;
	format_t *format;
	int sysctl_mib[4];
	const char *units;
} cpu_t;
typedef struct info_t {
	battall_t *battall; /* global battery data */
	batt_t *bl; /* battery level */
	batt_t *bsc; /* battery state, charging */
	batt_t *bsd; /* battery state, discharging */
	batt_t *bsg; /* battery state, critical */
	batt_t *bsa; /* battery state, disconnected */
	batt_t *bsm; /* battery state, max */
	batt_t *bt; /* battery time */
	cpu_t *cl; /* CPU usage level */
	cpu_t *cc; /* CPU clock speed */
} info_t;
typedef struct Vlist_t {
	void (*pFunc)(info_t*);
	unsigned short prio;
	struct Vlist_t *next;
} Vlist_t;

size_t gen_size;
int i;
Vlist_t *Vlist = NULL;

void print_usage(void);
static output_t *setoutput (info_t *info, char **fmtstring);
static output_t *mk_msg_part(char *msg, size_t msgsize);
static output_t *mk_info_part(info_t *info, uint_fast8_t infopart_type);
static void Vlist_register (void (*pFunc)(info_t*), uint_fast8_t prio);
void info_init_cl(info_t *info);
void info_init_cc(info_t *info);
void info_init_bl(info_t *info);
void info_init_bsc(info_t *info);
void info_init_bsd(info_t *info);
void info_init_bsa(info_t *info);
void info_init_bsg(info_t *info);
void info_init_bsm(info_t *info);
void info_init_bt(info_t *info);
void info_init_battall(info_t *info);
void info_ioctl_bif(info_t *info);
void info_ioctl_bio(info_t *info);
void info_get_cl(info_t *info);
void info_get_cc(info_t *info);
void info_get_bl(info_t *info);
void info_get_bsc(info_t *info);
void info_get_bsd(info_t *info);
void info_get_bsg(info_t *info);
void info_get_bsa(info_t *info);
void info_get_bsm(info_t *info);
void info_get_bs(info_t *info, uint_fast8_t state);
void info_get_bt(info_t *info);

int
main(int argc, char **argv) {
	unsigned int	waittime = 1;
#ifdef WITH_DWM
	uint_fast8_t	dwmmode = 0;
	Display		*xdpy = NULL;
	int		xscreen;
	Window		xroot;
#endif
	output_t	*output = NULL;
	output_t	*ostart = NULL;
	info_t 		info;
	Vlist_t		*vstart;
	char		*statstr;
	size_t		statstrsz;

	memset(&info, 0, sizeof(info_t));

	while ((i = getopt(argc, argv, "hxw:f:t:")) != -1) {
		switch(i) {
		case 'w':
			waittime = atoi(optarg);
			if (waittime < 1)
				errx(1, "wait time is < 1");
			break;
		case 'f':
			if (strlen(optarg) >= FORMAT_MAX)
				errx(1, "format string length >%d", FORMAT_MAX);
			statstrsz = sizeof(char) * (strlen(optarg)*2);
			statstr = malloc(statstrsz);
			memset(statstr, 0, statstrsz);
			output = setoutput(&info, &optarg);
			break;
		case 'x':
#ifdef WITH_DWM
			dwmmode = 1;
#endif
			break;
		case 'h':
		default:
			print_usage();
			return 1;
		}
	}
	
	if (!Vlist) {
		print_usage();
		return 1;
	}
	
	/* Our little bootstrap */
	vstart = Vlist;
	ostart = output;
	
	while (1) {
		while (Vlist) {
			(*(Vlist->pFunc))(&info);
			Vlist = Vlist->next;
		}
		Vlist = vstart;
		while (output) {
			while (output && output->conditional) {
				if (output->msgpart) output = output->next;
				else output = output->jump;
			}
			if (!output) break;
			if (output->msgpart) {
				strcat(statstr, output->msgpart);
			}
			output = output->next;
		}
#ifdef WITH_DWM
		if (dwmmode) {
			xdpy = XOpenDisplay(NULL);
			if (!xdpy) {
				errx(1, "Unable to open display '%s'",
					XDisplayName(NULL));
			}
			xscreen = DefaultScreen(xdpy);
			xroot = RootWindow(xdpy, xscreen);
			XStoreName(xdpy, xroot, statstr);
			XCloseDisplay(xdpy);
		} else
#endif
		{
			printf("%s\n", statstr);
			fflush(stdout);
		}
		memset(statstr, 0, statstrsz);
		output = ostart;

		sleep(waittime);
	}

	exit(0);
}

void
print_usage (void) {
#ifdef WITH_DWM
	printf("Usage: statf [-h] [-x] -f <format> -w <seconds>\n");
#else
	printf("Usage: statf [-h] -f <format> -w <seconds>\n");
#endif
}

static output_t *
setoutput (info_t *info, char **fmtstring) {
	output_t	*output = NULL;
	output_t	*noutput = NULL;
	output_t	*poutput = NULL;
	output_t	*coutput = NULL;
	char 		*s, *c;
	uint_fast8_t	fstring_state = FSTRING_STATE_ROOT;
	
	s = c = *fmtstring;
	
	while (s) {
		switch (fstring_state) {
		case FSTRING_STATE_ROOT:
			switch (*c) {
			case '%':
				if (c > s) {
					/* encountered a format string with text before. */
					gen_size = sizeof(char) * (c-s+1);
					/* malloc a buffer, copy the previous text to the working output part.
					 * make a new working output part */
					noutput = mk_msg_part(s, gen_size);
					/* set the start position of next part */
					s = c;
				}
				if (noutput && coutput) {
					coutput->jump = noutput;
					coutput = NULL;
				}
				fstring_state = FSTRING_STATE_FMT;
				break;
			case '{':
				if (c > s || coutput || !output) errx(1, "Invalid format string");
				++c;
				noutput = setoutput(info, &c);
				coutput = poutput;
				coutput->conditional = 1;
				while (noutput->next) {
					poutput->next = noutput;
					poutput = noutput;
					noutput = noutput->next;
				}
				if (*c == '\0') s = NULL;
				else s = c+1;
				break;
			case '}':
			case '\0':
				if (c > s) {
					gen_size = sizeof(char) * (c-s+1);
					noutput = mk_msg_part(s, gen_size);
				}
				if (noutput && coutput) {
					coutput->jump = noutput;
					coutput = NULL;
				}
				s = NULL;
				break;
			}
			break;
		case FSTRING_STATE_FMT:
			switch (*c) {
			case 'b':
				/* battery */
				++c;
				switch (*c) {
				case 'l': noutput = mk_info_part(info, INFO_SET_BL); break; /* battery level */
				case 't': noutput = mk_info_part(info, INFO_SET_BT); break; /* battery time */
				case 's':
					/* battery state */
					++c;
					switch (*c) {
						case 'c': noutput = mk_info_part(info, INFO_SET_BSC); break; /* charging */
						case 'd': noutput = mk_info_part(info, INFO_SET_BSD); break; /* discharging */
						case 'D': noutput = mk_info_part(info, INFO_SET_BSG); break; /* disconnected */
						case 'C': noutput = mk_info_part(info, INFO_SET_BSA); break; /* critical */
						case 'm': noutput = mk_info_part(info, INFO_SET_BSM); break; /* max */
					}
					break;
				}
				break;
			case 'c':
				/* CPU */
				++c;
				switch (*c) {
					case 'l': noutput = mk_info_part(info, INFO_SET_CL); break; /* CPU level */
					case 'c': noutput = mk_info_part(info, INFO_SET_CC); break; /* CPU clock */
				}
				break;
			}
			if (!noutput) errx(1, "Invalid format string");
			if (coutput) {
				coutput->jump = noutput;
				coutput = NULL;
			}
			s = c+1;
			fstring_state = FSTRING_STATE_ROOT;
			break;
		}
		++c;
		if (noutput) {
			if (poutput) {
				poutput->next = noutput;
				poutput = noutput;
			} else {
				output = poutput = noutput;
			}
			noutput = NULL;
		}
	}
	
	*fmtstring = c-1;
	return output;
}

static output_t *
mk_msg_part(char *msg, size_t msgsize) {
	output_t	*output = NULL;
	
	if (msg && msgsize) {
		output = malloc(sizeof(output_t));
		memset(output, 0, sizeof(output_t));
		output->msgpart = malloc(msgsize);
		strlcpy(output->msgpart, msg, msgsize);
	}
	
	return output;
}

output_t *
mk_info_part(info_t *info, uint_fast8_t infopart_type) {
	output_t	*output;
	format_t	**format;
	int		formatc;
	void		(*initpFunc)(info_t*);
	
	switch (infopart_type) {
	case INFO_SET_BL:
		if (!info->bl) {
			info->bl = malloc(sizeof(batt_t));
			memset(info->bl, 0, sizeof(batt_t));
			info->bl->units = "%";
		}
		format = &(info->bl->format);
		formatc = ++info->bl->formatc;
		initpFunc = &info_init_bl;
		break;
	case INFO_SET_BSC:
		if (!info->bsc) {
			info->bsc = malloc(sizeof(batt_t));
			memset(info->bsc, 0, sizeof(batt_t));
		}
		format = &(info->bsc->format);
		formatc = ++info->bsc->formatc;
		initpFunc = &info_init_bsc;
		break;
	case INFO_SET_BSD:
		if (!info->bsd) {
			info->bsd = malloc(sizeof(batt_t));
			memset(info->bsd, 0, sizeof(batt_t));
		}
		format = &(info->bsd->format);
		formatc = ++info->bsd->formatc;
		initpFunc = &info_init_bsd;
		break;
	case INFO_SET_BSG:
		if (!info->bsg) {
			info->bsg = malloc(sizeof(batt_t));
			memset(info->bsg, 0, sizeof(batt_t));
		}
		format = &(info->bsg->format);
		formatc = ++info->bsg->formatc;
		initpFunc = &info_init_bsg;
		break;
	case INFO_SET_BSA:
		if (!info->bsa) {
			info->bsa = malloc(sizeof(batt_t));
			memset(info->bsa, 0, sizeof(batt_t));
		}
		format = &(info->bsa->format);
		formatc = ++info->bsa->formatc;
		initpFunc = &info_init_bsa;
		break;
	case INFO_SET_BSM:
		if (!info->bsm) {
			info->bsm = malloc(sizeof(batt_t));
			memset(info->bsm, 0, sizeof(batt_t));
		}
		format = &(info->bsm->format);
		formatc = ++info->bsm->formatc;
		initpFunc = &info_init_bsm;
		break;
	case INFO_SET_BT:
		if (!info->bt) {
			info->bt = malloc(sizeof(batt_t));
			memset(info->bt, 0, sizeof(batt_t));
		}
		format = &(info->bt->format);
		formatc = ++info->bt->formatc;
		initpFunc = &info_init_bt;
		break;
	case INFO_SET_CL:
		if (!info->cl) {
			info->cl = malloc(sizeof(cpu_t));
			memset(info->cl, 0, sizeof(cpu_t));
			info->cl->units = "%";
		}
		format = &(info->cl->format);
		formatc = ++info->cl->formatc;
		initpFunc = &info_init_cl;
		break;
	case INFO_SET_CC:
		if (!info->cc) {
			gen_size = sizeof(cpu_t);
			info->cc = malloc(gen_size);
			memset(info->cc, 0, gen_size);
			info->cc->units = "MHz";
		}
		format = &(info->cc->format);
		formatc = ++info->cc->formatc;
		initpFunc = &info_init_cc;
		break;
	default:
		return NULL;
	}
	
	output = malloc(sizeof(output_t));
	memset(output, 0, sizeof(output_t));
	gen_size = sizeof(format_t) * formatc;
	*format = realloc(*format, gen_size); /* alloc or resize, depending on formatc */
	memset(*format+(formatc-1), 0, sizeof(format_t)); /* zero the last element */
	(*format+(formatc-1))->output = output; /* attach an output to last element */
	(*initpFunc)(info);
	
	return output;
}

static void
Vlist_register (void (*pFunc)(info_t*), uint_fast8_t prio) {
	Vlist_t **p, *ins;
	
	/* Registering a pFunc means inserting it into the list in order of
	 * prio (priority), while also ensuring there are no dupe pFuncs
	 * of the same priority. */
	
	for (p = &Vlist; *p != NULL; p = &(*p)->next) {
		if ((*p)->prio == prio && (*p)->pFunc == pFunc) return; /* dupe! */
		if ((*p)->prio > prio) break; /* insertion point found */
	}
	
	ins = malloc(sizeof(Vlist_t));
	ins->prio = prio;
	ins->pFunc = pFunc;

	ins->next = *p;
	*p = ins;
}

void
info_init_cl(info_t *info) {
	format_t	*format;
	/* CPU level initialisation */
	if (info->cl->formatc == 1) {
		gen_size = sizeof(int) * 2;
		if (sysctlnametomib("kern.cp_time", info->cl->sysctl_mib, &gen_size) != 0)
			errx(1, "sysctlnametomib(kern.cp_time) failed: (%d) %s", errno, strerror(errno));
		Vlist_register(&info_get_cl, VLIST_PRIO_GEN);
		info->cl->format[0].output->msgpart = malloc(sizeof(char) * 6);
	}
	format = &(info->cl->format[info->cl->formatc-1]);
	format->output->msgpart = info->cl->format[0].output->msgpart;
}

void
info_init_cc(info_t *info) {
	format_t	*format;
	/* CPU clock speed initialisation */
	if (info->cc->formatc == 1) {
		gen_size = sizeof(int) * 4;
		if (sysctlnametomib("dev.cpu.0.freq", info->cc->sysctl_mib, &gen_size) != 0)
			errx(1, "sysctlnametomib(dev.cpu.0.freq) failed: (%d) %s", errno, strerror(errno));
		Vlist_register(&info_get_cc, VLIST_PRIO_GEN);
		info->cc->format[0].output->msgpart = malloc(sizeof(char) * 10);
	}
	format = &(info->cc->format[info->cc->formatc-1]);
	format->output->msgpart = info->cc->format[0].output->msgpart;
}

void
info_init_bl(info_t *info) {
	/* battery level initialisation */
	if (info->bl->formatc == 1) {
		info_init_battall(info);
		Vlist_register(&info_get_bl, VLIST_PRIO_BIOREAD);
	}
}

void
info_init_bsc(info_t *info) {
	/* battery state (charging) initialisation */
	if (info->bsc->formatc == 1) {
		info_init_battall(info);
		Vlist_register(&info_get_bsc, VLIST_PRIO_BIOREAD);
	}
}

void
info_init_bsd(info_t *info) {
	/* battery state (discharging) initialisation */
	if (info->bsd->formatc == 1) {
		info_init_battall(info);
		Vlist_register(&info_get_bsd, VLIST_PRIO_BIOREAD);
	}
}

void
info_init_bsg(info_t *info) {
	/* battery state (disconnected) initialisation */
	if (info->bsg->formatc == 1) {
		info_init_battall(info);
		Vlist_register(&info_get_bsg, VLIST_PRIO_BIOREAD);
	}
}

void
info_init_bsa(info_t *info) {
	/* battery state (critical) initialisation */
	if (info->bsa->formatc == 1) {
		info_init_battall(info);
		Vlist_register(&info_get_bsa, VLIST_PRIO_BIOREAD);
	}
}

void
info_init_bsm(info_t *info) {
	/* battery state (max) initialisation */
	if (info->bsm->formatc == 1) {
		info_init_battall(info);
		Vlist_register(&info_get_bsm, VLIST_PRIO_BIOREAD);
	}
}

void
info_init_bt(info_t *info) {
	/* battery state (max) initialisation */
	if (info->bt->formatc == 1) {
		info_init_battall(info);
		Vlist_register(&info_get_bt, VLIST_PRIO_BIOREAD);
	}
}

void
info_init_battall(info_t *info) {
	Vlist_register(&info_ioctl_bif, VLIST_PRIO_ACPIIOCTL);
	Vlist_register(&info_ioctl_bio, VLIST_PRIO_ACPIIOCTL);
	if (!info->battall) {
		info->battall = malloc(sizeof(battall_t));
		info->battall->acpifd = open(ACPIDEV, O_RDONLY);
		if (info->battall->acpifd == -1)
			err(EX_OSFILE, ACPIDEV);
	}
}

void
info_ioctl_bif(info_t *info) {
	battall_t	*battall = info->battall;

	battall->battio.unit = 0;
	if (ioctl(battall->acpifd, ACPIIO_BATT_GET_BIF, &(battall->battio)) == -1)
		err(EX_IOERR, "get battery info failed");
	if (battall->battio.bif.units == 0)
		battall->pwr_units = "mW";
	else
		battall->pwr_units = "mA";
}

void
info_ioctl_bio(info_t *info) {
	battall_t	*battall = info->battall;
	
	battall->battio.unit = 0;
	if (ioctl(battall->acpifd, ACPIIO_BATT_GET_BATTINFO, &(battall->battio)) == -1)
		err(EX_IOERR, "get charge status failed");
}

void
info_get_cl(info_t *info) {
	static long	cptime_cur[CPUSTATES];
	static long	cptime_lst[CPUSTATES];
	double		cptotal = 0.0;
	long		cptime[CPUSTATES];
	long		clpct = 0;
	cpu_t		*cl = info->cl;
	
	gen_size = sizeof(cptime_cur);
	if (sysctl(cl->sysctl_mib, 2, &cptime_cur, &gen_size, NULL, (size_t) 0) == -1)
		errx(1, "sysctl(kern.cp_time) failed: (%d) %s", errno, strerror(errno));
	
	for (i = 0; i < CPUSTATES; i++) {
		cptime[i] = cptime_cur[i] - cptime_lst[i];
		cptotal += cptime[i];
		cptime_lst[i] = cptime_cur[i];
	}
	
	if (cptotal > 0) {
		clpct = lround(100.0 - (100.0 * cptime[4] / cptotal));
	}
	
	sprintf(cl->format[0].output->msgpart, "%ld%s", clpct, cl->units);
}

void
info_get_cc(info_t *info) {
	int		cpclock = 0;
	cpu_t		*cc = info->cc;
	
	gen_size = sizeof(int);
	if (sysctl(cc->sysctl_mib, 4, &cpclock, &gen_size, NULL, (size_t) 0) == -1)
		errx(1, "sysctl(dev.cpu.0.freq) failed: (%d) %s", errno, strerror(errno));
	
	sprintf(cc->format[0].output->msgpart, "%d%s", cpclock, cc->units);
}

void
info_get_bl(info_t *info) {
	batt_t		*bl = info->bl;
	battall_t	*battall = info->battall;
	char		**msgpart = &(bl->format[0].output->msgpart);
	
	if (battall->battio.battinfo.state == ACPI_BATT_STAT_NOT_PRESENT ||
	battall->battio.battinfo.cap == -1) {
		if (*msgpart) {
			free(*msgpart);
			for (i = 0; i < bl->formatc; ++i) {
				bl->format[i].output->msgpart = NULL;
			}
		}
		return;
	}
	
	if (!*msgpart) *msgpart = malloc(sizeof(char) * 4);
	for (i = 0; i < bl->formatc; ++i) {
		bl->format[i].output->msgpart = *msgpart;
	}
	sprintf(*msgpart, "%d%s", battall->battio.battinfo.cap, bl->units);
}

void
info_get_bsc(info_t *info) {
	info_get_bs(info, BSTATE_CHARGING);
}
void
info_get_bsd(info_t *info) {
	info_get_bs(info, BSTATE_DISCHARGING);
}
void
info_get_bsa(info_t *info) {
	info_get_bs(info, BSTATE_CRITICAL);
}
void
info_get_bsg(info_t *info) {
	info_get_bs(info, BSTATE_DISCONNECTED);
}
void
info_get_bsm(info_t *info) {
	info_get_bs(info, BSTATE_MAX);
}
void
info_get_bs(info_t *info, uint_fast8_t state) {
	batt_t		*bs;
	battall_t	*battall = info->battall;
	uint_fast8_t	test;

	switch (state) {
	case BSTATE_CHARGING:
		bs = info->bsc;
		test = (battall->battio.battinfo.state & ACPI_BATT_STAT_CHARGING) > 0;
		break;
	case BSTATE_DISCHARGING:
		bs = info->bsd;
		test = (battall->battio.battinfo.state & ACPI_BATT_STAT_DISCHARG) > 0;
		break;
	case BSTATE_CRITICAL:
		bs = info->bsa;
		test = (battall->battio.battinfo.state & ACPI_BATT_STAT_CRITICAL) > 0;
		break;
	case BSTATE_DISCONNECTED:
		bs = info->bsg;
		test = (battall->battio.battinfo.state == ACPI_BATT_STAT_NOT_PRESENT);
		break;
	case BSTATE_MAX:
		bs = info->bsm;
		test = (battall->battio.battinfo.state == 0);
		break;
	}
	
	if (test) {
		if (!bs->format[0].output->msgpart) {
			for (i = 0; i < bs->formatc; ++i) {
				bs->format[i].output->msgpart = (char *) 1; /* naughty? */
			}
		}
		return;
	}
	
	if (bs->format[0].output->msgpart) {
		for (i = 0; i < bs->formatc; ++i) {
			bs->format[i].output->msgpart = NULL;
		}
	}
}

void
info_get_bt(info_t *info) {
	batt_t		*bt = info->bt;
	battall_t	*battall = info->battall;
	char		**msgpart = &(bt->format[0].output->msgpart);
	uint_fast8_t	hours, mins;
	
	if (battall->battio.battinfo.min == -1) {
		if (*msgpart) {
			free(*msgpart);
			for (i = 0; i < bt->formatc; ++i) {
				bt->format[i].output->msgpart = NULL;
			}
		}
		return;
	}
	
	if (!*msgpart) *msgpart = malloc(sizeof(char) * 6);
	for (i = 0; i < bt->formatc; ++i) {
		bt->format[i].output->msgpart = *msgpart;
	}
	hours = battall->battio.battinfo.min / 60;
	mins = battall->battio.battinfo.min % 60;
	sprintf(*msgpart, "%uh%02um", hours, mins);
}
